﻿using AztecCodeReaderWindows10;
using System;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using Windows.ApplicationModel;
using Windows.Devices.Enumeration;
using Windows.Graphics.Display;
using Windows.Graphics.Imaging;
using Windows.Media;
using Windows.Media.Capture;
using Windows.Media.Core;
using Windows.Media.Devices;
using Windows.Media.MediaProperties;
using Windows.System.Display;
using Windows.System.Profile;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Navigation;


namespace AztecDemoWin10
{
    [ComImport]
    [Guid("5b0d3235-4dba-4d44-865e-8f1d0e4fd04d")]
    [InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    unsafe interface IMemoryBufferByteAccess
    {
        void GetBuffer(out byte* buffer, out uint capacity);
    }

    public sealed partial class MainPage : Page
    {
        private readonly DisplayInformation _displayInformation = DisplayInformation.GetForCurrentView();
        private DisplayOrientations _displayOrientation = DisplayOrientations.Portrait;

        // Rotation metadata to apply to the preview stream (MF_MT_VIDEO_ROTATION)
        // Reference: http://msdn.microsoft.com/en-us/library/windows/apps/xaml/hh868174.aspx
        private static readonly Guid RotationKey = new Guid("C380465D-2271-428C-9B83-ECEA3B4A85C1");

        private readonly DisplayRequest _displayRequest = new DisplayRequest();
        private readonly SystemMediaTransportControls _systemMediaControls = SystemMediaTransportControls.GetForCurrentView();

        private bool _isInitialized = false;
        private bool _isPreviewing = false;
        private bool _isRecognizing = false;

        private const int MAX_RESOLUTION = 640 * 480;

        private DispatcherTimer _focusTimer;
        private MediaCapture _mediaCapture;
        private FocusControl _focusControl;
        private SceneAnalysisEffect _sceneAnalysisEffect;
        private VideoEncodingProperties _previewProperties;
        private AztecReader _aztecReader;
        public MainPage()
        {
            this.InitializeComponent();

            NavigationCacheMode = NavigationCacheMode.Required;
            DisplayInformation.AutoRotationPreferences = DisplayOrientations.Portrait;

            Application.Current.Suspending += Application_Suspending;
            Application.Current.Resuming += Application_Resuming;
            _aztecReader = new AztecReader();
        }

        private void Page_Loaded(object sender, RoutedEventArgs e)
        {
            var size = ActualWidth > ActualHeight ? ActualHeight : ActualWidth;
            rectangle.Width = rectangle.Height = size - 64;
        }

        private async void Application_Suspending(object sender, SuspendingEventArgs e)
        {
            if (Frame.CurrentSourcePageType == typeof(MainPage))
            {
                var deferral = e.SuspendingOperation.GetDeferral();

                await CleanupCameraAsync();
                _displayInformation.OrientationChanged -= DisplayInformation_OrientationChanged;

                deferral.Complete();
            }
        }

        private async void Application_Resuming(object sender, object o)
        {
            if (Frame.CurrentSourcePageType == typeof(MainPage))
            {
                _displayOrientation = _displayInformation.CurrentOrientation;
                _displayInformation.OrientationChanged += DisplayInformation_OrientationChanged;

                await InitializeCameraAsync();
            }
        }

        protected override async void OnNavigatedTo(NavigationEventArgs e)
        {
            _displayOrientation = _displayInformation.CurrentOrientation;
            _displayInformation.OrientationChanged += DisplayInformation_OrientationChanged;

            await InitializeCameraAsync();
        }

        protected override async void OnNavigatingFrom(NavigatingCancelEventArgs e)
        {
            await CleanupCameraAsync();
            _displayInformation.OrientationChanged -= DisplayInformation_OrientationChanged;
        }

        private async void DisplayInformation_OrientationChanged(DisplayInformation sender, object args)
        {
            _displayOrientation = sender.CurrentOrientation;
            if (_isPreviewing)
            {
                await SetPreviewRotationAsync();
            }
        }

        private async void MediaCapture_Failed(MediaCapture sender, MediaCaptureFailedEventArgs errorEventArgs)
        {
            Debug.WriteLine("MediaCapture_Failed: (0x{0:X}) {1}", errorEventArgs.Code, errorEventArgs.Message);
            await CleanupCameraAsync();
        }

        private async Task InitializeCameraAsync()
        {
            Debug.WriteLine("InitializeCameraAsync");

            if (_mediaCapture == null)
            {
                var cameraDevice = await FindCameraDeviceByPanelAsync(Windows.Devices.Enumeration.Panel.Back);

                if (cameraDevice == null)
                {
                    cameraDevice = await FindCameraDeviceByPanelAsync(Windows.Devices.Enumeration.Panel.Front);
                }
                if (cameraDevice == null)
                {
                    Debug.WriteLine("No camera device found!");
                    return;
                }

                _mediaCapture = new MediaCapture();
                _mediaCapture.Failed += MediaCapture_Failed;

                var settings = new MediaCaptureInitializationSettings { VideoDeviceId = cameraDevice.Id };
                try
                {
                    await _mediaCapture.InitializeAsync(settings);
                    _isInitialized = true;
                }
                catch (UnauthorizedAccessException)
                {
                    Debug.WriteLine("The app was denied access to the camera");
                }

                _focusControl = _mediaCapture.VideoDeviceController.FocusControl;
                if (_focusControl.Supported)
                {
                    _focusControl.Configure(new FocusSettings { Mode = FocusMode.Single });

                    _focusTimer = new DispatcherTimer();
                    _focusTimer.Tick += onFocusTimer;
                    _focusTimer.Interval = TimeSpan.FromSeconds(3.0);
                    _focusTimer.Start();
                }

                var videoRes = _mediaCapture.VideoDeviceController.GetAvailableMediaStreamProperties(MediaStreamType.VideoPreview).Where(x => ((VideoEncodingProperties)x).Height * ((VideoEncodingProperties)x).Width <= MAX_RESOLUTION).OrderByDescending(x => ((VideoEncodingProperties)x).Height * ((VideoEncodingProperties)x).Width).FirstOrDefault();
                await _mediaCapture.VideoDeviceController.SetMediaStreamPropertiesAsync(MediaStreamType.VideoPreview, videoRes);
                _previewProperties = (VideoEncodingProperties)videoRes;

                var definition = new SceneAnalysisEffectDefinition();
                _sceneAnalysisEffect = (SceneAnalysisEffect)await _mediaCapture.AddVideoEffectAsync(definition, MediaStreamType.VideoPreview);
                _sceneAnalysisEffect.SceneAnalyzed += SceneAnalysisEffect_SceneAnalyzed;

                if (_isInitialized)
                {
                    await StartPreviewAsync();
                }
            }
        }
        private async void SceneAnalysisEffect_SceneAnalyzed(SceneAnalysisEffect sender, SceneAnalyzedEventArgs args)
        {
            if (!_isRecognizing)
            {
                _isRecognizing = true;

                if (_isPreviewing)
                {
                    try
                    {
                        var previewFrame = await GetPreviewFrameAsSoftwareBitmapAsync();
                        Recognize(previewFrame);
                        _isRecognizing = false;

                    }
                    catch (Exception ex)
                    {
                        Debug.WriteLine(ex);
                    }
                }
            }
        }

        private unsafe void Recognize(SoftwareBitmap bitmap)
        {
            try
            {
                if (bitmap.BitmapPixelFormat == BitmapPixelFormat.Nv12)
                {
                    using (var buffer = bitmap.LockBuffer(BitmapBufferAccessMode.ReadWrite))
                    using (var reference = buffer.CreateReference())
                    {
                        if (reference is IMemoryBufferByteAccess)
                        {
                            byte* data;
                            uint capacity;
                            ((IMemoryBufferByteAccess)reference).GetBuffer(out data, out capacity);

                            var desc = buffer.GetPlaneDescription(0);
                            try
                            {

                                byte[] arr = new byte[capacity];
                                Marshal.Copy(new IntPtr(data), arr, 0, (int)capacity);
                                string result = _aztecReader.Read(arr, desc.Width, desc.Height);

                                if (!string.IsNullOrEmpty(result))
                                {
                                    Windows.ApplicationModel.Core.CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(CoreDispatcherPriority.High, () =>
                                    {
                                        tbResult.Text = result;
                                        gridResult.Visibility = Visibility.Visible;
                                    }).AsTask();
                                    Debug.WriteLine(result);
                                }
                                else
                                {
                                    Debug.WriteLine("not recognized");
                                }
                            }
                            catch (Exception ex)
                            {
                                Debug.WriteLine(ex);
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex);
            }
        }

        private void onFocusTimer(object sender, object e)
        {
            try
            {
                if (_focusControl.Supported && _focusControl.FocusState != MediaCaptureFocusState.Searching)
                {
                    _focusControl.FocusAsync();
                }
            }
            catch (Exception ex)
            {
                _focusTimer.Stop();
            }
        }

        private async Task StartPreviewAsync()
        {
            Debug.WriteLine("StartPreviewAsync");
            _displayRequest.RequestActive();
            PreviewControl.Source = _mediaCapture;

            await _mediaCapture.StartPreviewAsync();
            _isPreviewing = true;

            if (_isPreviewing)
            {
                await SetPreviewRotationAsync();
            }
        }

        private async Task SetPreviewRotationAsync()
        {
            int rotationDegrees = ConvertDisplayOrientationToDegrees(_displayOrientation);

            var props = _mediaCapture.VideoDeviceController.GetMediaStreamProperties(MediaStreamType.VideoPreview);
            props.Properties.Add(RotationKey, rotationDegrees);
            await _mediaCapture.SetEncodingPropertiesAsync(MediaStreamType.VideoPreview, props, null);
        }

        private async Task StopPreviewAsync()
        {
            _isPreviewing = false;
            await _mediaCapture.StopPreviewAsync();

            await Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
            {
                PreviewControl.Source = null;

                _displayRequest.RequestRelease();
            });
        }

        private async Task<SoftwareBitmap> GetPreviewFrameAsSoftwareBitmapAsync()
        {
            try
            {
                var videoFrame = new VideoFrame(BitmapPixelFormat.Nv12, (int)_previewProperties.Width, (int)_previewProperties.Height);
                using (var currentFrame = await _mediaCapture.GetPreviewFrameAsync(videoFrame))
                {
                    return currentFrame.SoftwareBitmap;
                }
            }
            catch (Exception ex)
            {
                throw;
            }
        }

        private async Task CleanupCameraAsync()
        {
            if (_isInitialized)
            {
                if (_isPreviewing)
                {
                    await StopPreviewAsync();
                }

                _isInitialized = false;
            }

            if (_focusTimer != null && _focusTimer.IsEnabled)
            {
                _focusTimer.Stop();
            }

            if (_mediaCapture != null)
            {
                if (_sceneAnalysisEffect != null)
                {
                    _sceneAnalysisEffect.SceneAnalyzed -= SceneAnalysisEffect_SceneAnalyzed;
                    await _mediaCapture.ClearEffectsAsync(MediaStreamType.VideoPreview);
                    _sceneAnalysisEffect = null;
                }
                _mediaCapture.Failed -= MediaCapture_Failed;
                _mediaCapture.Dispose();
                _mediaCapture = null;
            }
        }

        private async Task<DeviceInformation> FindCameraDeviceByPanelAsync(Windows.Devices.Enumeration.Panel desiredPanel)
        {
            var allVideoDevices = await DeviceInformation.FindAllAsync(DeviceClass.VideoCapture);

            DeviceInformation desiredDevice = allVideoDevices.FirstOrDefault(x => x.EnclosureLocation != null && x.EnclosureLocation.Panel == desiredPanel);

            return desiredDevice ?? allVideoDevices.FirstOrDefault();
        }

        private int ConvertDisplayOrientationToDegrees(DisplayOrientations orientation)
        {
            switch (orientation)
            {
                case DisplayOrientations.Portrait:
                    return 90;
                case DisplayOrientations.LandscapeFlipped:
                    return 180;
                case DisplayOrientations.PortraitFlipped:
                    return 270;
                case DisplayOrientations.Landscape:
                default:
                    return 0;
            }
        }

        private async void Focus_Tapped(object sender, TappedRoutedEventArgs e)
        {
            if (_focusControl.Supported && _focusControl.FocusState != MediaCaptureFocusState.Searching)
            {
                await _focusControl.FocusAsync();
            }
        }

        private void Button_Click(object sender, RoutedEventArgs e)
        {
            gridResult.Visibility = Visibility.Collapsed;
            tbResult.Text = string.Empty;
        }
    }
}
