﻿using AztecCodeReaderWindows81;
using Lumia.Imaging;
using System;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using VideoEffects;
using Windows.ApplicationModel;
using Windows.Graphics.Display;
using Windows.Media.Capture;
using Windows.Media.Devices;
using Windows.Media.MediaProperties;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Navigation;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234238

namespace AztecDemoWindows81
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class MainPage : Page
    {
        private FocusControl _focusControl;
        private TorchControl _torchControl;
        private DispatcherTimer _focusTimer;
        private VideoRotation _rotation;

        private const int MAX_RESOLUTION = 640 * 480;

        private bool _isCameraStarted;

        private MediaCapture _mediaCapture;
        private AztecReader _aztecReader;

        public MainPage()
        {
            _aztecReader = new AztecReader();
            this.InitializeComponent();

            NavigationCacheMode = NavigationCacheMode.Required;
            DisplayInformation.AutoRotationPreferences = DisplayOrientations.Portrait;

            Application.Current.Suspending += Application_Suspending;
            Application.Current.Resuming += Application_Resuming;
        }

        private void Page_Loaded(object sender, RoutedEventArgs e)
        {
            var size = ActualWidth > ActualHeight ? ActualHeight : ActualWidth;
            rectangle.Width = rectangle.Height = size - 196;
        }

        protected async override void OnNavigatedTo(NavigationEventArgs e)
        {
            await InitializeAsync();
        }

        protected async override void OnNavigatedFrom(NavigationEventArgs e)
        {
            await Dispose();
        }

        private async void Application_Suspending(object sender, SuspendingEventArgs e)
        {
            if (Frame.CurrentSourcePageType == typeof(MainPage))
            {
                var deferral = e.SuspendingOperation.GetDeferral();

                await Dispose();

                deferral.Complete();
            }
        }

        private async void Application_Resuming(object sender, object o)
        {
            if (Frame.CurrentSourcePageType == typeof(MainPage))
            {
                await InitializeAsync();
            }
        }

        public async Task InitializeAsync()
        {
            try
            {
                BtnFlashLight.IsEnabled = false;

                _mediaCapture = new MediaCapture();

                var camera = await CameraHelper.GetCameraID(Windows.Devices.Enumeration.Panel.Back);
                if(camera == null)
                {
                    camera = await CameraHelper.GetCameraID(Windows.Devices.Enumeration.Panel.Front);
                }
                if(camera == null)
                {
                    return;
                }

                var cameraSettings = new MediaCaptureInitializationSettings()
                {
                    StreamingCaptureMode = StreamingCaptureMode.Video,
                    VideoDeviceId = camera.Id
                };

                await _mediaCapture.InitializeAsync(cameraSettings);

                var videoRes = _mediaCapture.VideoDeviceController.GetAvailableMediaStreamProperties(MediaStreamType.VideoPreview).FirstOrDefault();
                foreach (var res in _mediaCapture.VideoDeviceController.GetAvailableMediaStreamProperties(MediaStreamType.VideoPreview))
                {
                    VideoEncodingProperties resolution = (VideoEncodingProperties)res;
                    Debug.WriteLine(resolution.Width + " " + resolution.Height);
                    var currentResolution = resolution.Width * resolution.Height;
                    if (currentResolution <= MAX_RESOLUTION)
                    {
                        videoRes = resolution;
                        break;
                    }

                }

                var definition = new LumiaAnalyzerDefinition(ColorMode.Yuv420Sp, ((VideoEncodingProperties)videoRes).Width, AnalyzeBitmap);
                await _mediaCapture.AddEffectAsync(MediaStreamType.VideoPreview, definition.ActivatableClassId, definition.Properties);

                await _mediaCapture.VideoDeviceController.SetMediaStreamPropertiesAsync(MediaStreamType.VideoPreview, videoRes);

                DisplayInformation displayInfo = DisplayInformation.GetForCurrentView();
                displayInfo.OrientationChanged += DisplayInfo_OrientationChanged;
                DisplayInfo_OrientationChanged(displayInfo, null);

                _focusControl = _mediaCapture.VideoDeviceController.FocusControl;
                _torchControl = _mediaCapture.VideoDeviceController.TorchControl;

                BtnFlashLight.IsChecked = true;
                BtnFlashLight.IsEnabled = _torchControl.Supported;

                if (_torchControl.Supported)
                {
                    _torchControl.Enabled = false;
                }

                if (_focusControl.Supported)
                {
                    _focusTimer = new DispatcherTimer();
                    _focusTimer.Tick += onFocusTimer;
                    _focusTimer.Interval = TimeSpan.FromSeconds(3.0);
                    _focusTimer.Start();
                }

                captureElement.Source = _mediaCapture;

                await _mediaCapture.StartPreviewAsync();
                _isCameraStarted = true;
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Error InitializeAsync Camera: {0}", ex);
            }
        }

        private void onFocusTimer(object sender, object e)
        {
            try
            {
                if (_focusControl.Supported)
                {
                    _focusControl.FocusAsync();
                }
            }
            catch (Exception ex)
            {
                _focusTimer.Stop();
            }
        }

        private void AnalyzeBitmap(Bitmap bitmap, TimeSpan time)
        {
            try
            {
                Debug.WriteLine("not recognized");
                var result = _aztecReader.Read(bitmap.Buffers[0].Buffer.ToArray(), (int)bitmap.Dimensions.Width, (int)bitmap.Dimensions.Height);
                if (!string.IsNullOrEmpty(result))
                {
                    Windows.ApplicationModel.Core.CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(CoreDispatcherPriority.High, () =>
                    {
                        tbResult.Text = result;
                        gridResult.Visibility = Visibility.Visible;
                    }).AsTask();
                    Debug.WriteLine(result);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex);
            }
        }

        private void DisplayInfo_OrientationChanged(DisplayInformation sender, object args)
        {
            if (_mediaCapture != null)
            {
                _rotation = CameraHelper.VideoRotationLookup(sender.CurrentOrientation, false);
                _mediaCapture.SetPreviewRotation(_rotation);
            }
        }

        public async Task Dispose()
        {
            try
            {
                if (_torchControl.Supported)
                {
                    _torchControl.Enabled = false;
                }
                if (_mediaCapture != null)
                {
                    if (_isCameraStarted)
                    {
                        await _mediaCapture.StopPreviewAsync();
                    }
                    _mediaCapture.Dispose();
                    _mediaCapture = null;
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine("Dispose Camera: {0}", ex);
            }
        }

        private async void Focus_Tapped(object sender, TappedRoutedEventArgs e)
        {
            try
            {
                if (_focusControl.Supported) 
                {
                    await _focusControl.FocusAsync();
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex);
            }
        }

        private void Button_Click(object sender, RoutedEventArgs e)
        {
            tbResult.Text = string.Empty;
            gridResult.Visibility = Visibility.Collapsed;

        }

        private void BtnInfo_Click(object sender, RoutedEventArgs e)
        {
            Frame.Navigate(typeof(InfoPage));
        }

        private void BtnFlashLight_Checked(object sender, RoutedEventArgs e)
        {
            if (_torchControl.Supported)
            {
                _torchControl.Enabled = false;
            }
        }

        private void BtnFlashLight_Unchecked(object sender, RoutedEventArgs e)
        {
            if (_torchControl.Supported)
            {
                _torchControl.Enabled = true;
            }
        }
    }
}
